// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "base.h"
#include "framebase.h"
#include "fitsimage.h"
#include "NaN.h"
#include "blt.h"

void FrameBase::bltProjection(char* xname, char* yname, 
			      char* xcname, char* ycname,
			      CoordSystem sys, SkyFrame sky,
			      const Vector& p1, const Vector& p2, 
			      double width, int avg)
{
  if (!currentFits)
    return;

  int num = (p2-p1).length() +1;

  double* x = (double*)malloc(num*sizeof(double));
  double* y = (double*)malloc(num*sizeof(double));
  double* xc = (double*)malloc(num*sizeof(double));
  double* yc = (double*)malloc(num*sizeof(double));

  bltProjFits(x, y, xc, yc, num, sys, sky, p1, p2, width, avg);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, y, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* xxc;
  if (Blt_GetVector(interp, xcname, &xxc) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xxc, xc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yyc;
  if (Blt_GetVector(interp, ycname, &yyc) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yyc, yc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  return;

 error:
  result = TCL_ERROR;
  return;
}

void FrameBase::bltProjFits(double* x, double* y, 
			    double* xc, double* yc, int num,
			    CoordSystem sys, SkyFrame sky,
			    const Vector& p1, const Vector& p2, 
			    double width, int avg)
{
  int* cnt = new int[num];
  Vector s = (p2-p1).normalize();

  // basics
  FitsImage* sptr = currentFits;
  int mosaic = isMosaic();

  // variable
  FitsBound* params = sptr->getDataParams(currentScale->scanMode());

  // main loop
  for (long ii=0; ii<num; ii++) {
    Vector t = p1 + s*ii;

    x[ii] = ii+1;

    if (mosaic) {
      sptr = currentFits;
      params = sptr->getDataParams(currentScale->scanMode());
    }

    do {
      Vector tv = sptr->mapFromRef(t,sys,sky);
      xc[ii] = tv[0];
      yc[ii] = tv[1];

      Vector z = t * sptr->getRefToData();

      if (z[0]>=params->xmin && z[0]<params->xmax && 
	  z[1]>=params->ymin && z[1]<params->ymax) {

	// check for nan
	double v = sptr->getValueDouble(z);
	if (!isnand(v)) {
	  y[ii] = v;
	  cnt[ii] = 1;
	}
	else
	  y[ii] = getnand();

	FitsImage* spptr = currentFits;
	FitsBound* pparams = spptr->getDataParams(currentScale->scanMode());

	Vector ss = Vector(-s[1],s[0]);

	for (long jj=1; jj<width; jj++) {
	  Vector tt = p1 + s*ii + ss*jj;

	  if (mosaic) {
	    spptr = currentFits;
	    pparams = spptr->getDataParams(currentScale->scanMode());
	  }

	  do {
	    Vector zz = tt * spptr->getRefToData();
		
	    if (zz[0]>=pparams->xmin && zz[0]<pparams->xmax && 
		zz[1]>=pparams->ymin && zz[1]<pparams->ymax) {

	      double vvalue = spptr->getValueDouble(zz);
	      // check for nan
	      if (!isnand(vvalue) && !isnand(y[ii])) {
		y[ii] += vvalue;
		cnt[ii]++;
	      }
	      else
		y[ii] = getnand();

	      break;
	    }
	    else
	      if (mosaic) {
		spptr = spptr->nextMosaic();
		if (spptr)
		  params = spptr->getDataParams(currentScale->scanMode());
	      }
	  }
	  while (mosaic && spptr);

	  if (!spptr)
	    y[ii] = getnand();
	}

	break;
      }
      else
	if (mosaic) {
	  sptr = sptr->nextMosaic();
	  if (sptr)
	    params = sptr->getDataParams(currentScale->scanMode());
	}
    }
    while (mosaic && sptr);

    if (!sptr)
      y[ii] = getnand();
  }

  // average if needed
  if (avg)
    for (long ii=0; ii<num; ii++)
      if (!isnand(y[ii]))
	y[ii] /= cnt[ii];

      // convert nand's to low
  for (int j=0; j<num; j++)
    if (isnand(y[j]))
      y[j] = currentScale->low();

  delete [] cnt;
}

void FrameBase::bltCut(char* xname, char* yname, 
		       Orientation axis, const Vector& rr)
{
  if (!currentFits)
    return;

  // Vector r is in Widget coords
  Vector r = rr;

  int size;
  if (axis == XX)
    size = options->width;
  else
    size = options->height;

  long length = (size+1) * 2;
  double* x = (double*)malloc(sizeof(double)*length);
  double* y = (double*)malloc(sizeof(double)*length);

  if (*currentMosaicCount == 0 && *currentSliceCount == 0) {
    for (int i=0; i<=size; i++) {
      x[i*2] = i;
      x[i*2+1] = i;
      y[i*2] = 0;
      y[i*2+1] = 0;
    }
  }
  else
    bltCutFits(x, y, size, r, axis);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK)
    goto error;

  if (Blt_ResetVector(yy, y, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  return;

 error:
    result = TCL_ERROR;
    return;
}

void FrameBase::bltCutFits(double* x, double* y, int size, Vector& r, 
			   Orientation axis)
{
  // basics
  FitsImage* sptr = currentFits;
  int mosaic = isMosaic();

  // variable
  FitsBound* params = sptr->getDataParams(currentScale->scanMode());
  double prev = currentScale->low();

  // main loop
  for (int ii=0; ii<=size; ii++) {
    double v = currentScale->low();

    if (mosaic) {
      sptr = currentFits;
      params = sptr->getDataParams(currentScale->scanMode());
    }

    do {
      Vector img;
      if (axis == XX)
	img = Vector(1+ii,r[1]) * sptr->getWidgetToData();
      else
	img = Vector(r[0],1+ii) * sptr->getWidgetToData();

      if (img[0]>=params->xmin && img[0]<params->xmax && 
	  img[1]>=params->ymin && img[1]<params->ymax) {
	double value = sptr->getValueDouble(img);

	if (!isnand(value))
	  v = value;

	break;
      }
      else {
	if (mosaic) {
	  sptr = sptr->nextMosaic();
	  if (sptr)
	    params = sptr->getDataParams(currentScale->scanMode());
	}
      }
    }
    while (mosaic && sptr);

    x[2*ii] = ii;
    x[2*ii +1] = ii;
    
    y[2*ii] = prev;
    y[2*ii +1] = v;
    prev = v;
  }
}
